/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvdta.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvip/qvipp/qvipp.h>
#include <qvdta/qvdta.h>

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

QVector< QVector< QPoint > > CountingSort(const QVImage<uChar, 1> &image)
	{
	QVector< QVector <QPoint> > result(256);
	const QVector<int> histogram = HistogramRange(image);

	for (int k=0; k<256; k++)
		result[k].reserve(histogram[k]);

	QVIMAGE_INIT_READ(uChar,image);
	for(uInt row = 0; row < image.getRows(); row++)
		for(uInt col = 0; col < image.getCols(); col++)
			result[QVIMAGE_PIXEL(image, col, row,0)].append(QPoint(col, row));
	
	return result;
	}


#include <qvmath/qvdisjointset.h>
#include <qvmath/qvvector.h>

QList<QPointF> GetMaximalResponsePoints1(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	const int sizeMax = 2;
	const int cols = cornerResponseImage.getCols(), rows = cornerResponseImage.getRows();
	QVImage<uChar> binaryResponseImage(cols, rows);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);
	QVIMAGE_INIT_WRITE(uChar,binaryResponseImage);

	// 1. Get image with local maximums
	Set(binaryResponseImage,0);
	sFloat actual;
	for(int row = sizeMax; row < rows-sizeMax; row++)
		for(int col = sizeMax; col < cols-sizeMax; col++)
			{
			actual = QVIMAGE_PIXEL(cornerResponseImage, col, row,0);
			if (actual >= threshold)
				{
				QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = IPP_MAX_8U;
				for (int j = row-sizeMax; (j < row+sizeMax) && (QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) > 0); j++)
					for (int i = col-sizeMax; i < col+sizeMax; i++)
						if ( ((i != col) || (j != row)) && (actual <= QVIMAGE_PIXEL(cornerResponseImage, i, j, 0)) )
							{
							QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = 0;
							break;
							}
				}
			}

	// 2. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	for(uInt row = 0; row < binaryResponseImage.getRows(); row++)
		for(uInt col = 0; col < binaryResponseImage.getCols(); col++)
			if (QVIMAGE_PIXEL(binaryResponseImage, col, row,0))
				sortedPoints.insertMulti(QVIMAGE_PIXEL(cornerResponseImage, col, row,0), QPointF(col+2, row+2));

	QList<QPointF> result;
	foreach(sFloat key, sortedPoints.uniqueKeys())
		result += sortedPoints.values(key);

	return result;
	}

QList<QPointF> GetMaximalResponsePoints3(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	static const int	xD[4] = { -1, +0, +1, +0 },
				yD[4] = { +0, -1, +0, +1 };

	const int rows = cornerResponseImage.getRows(), cols = cornerResponseImage.getCols();
	QVDisjointSet disjointSet(cornerResponseImage);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);

	// 1. Join in the disjoint set all the neighbour pixels
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row,0) >= threshold)
				{
				// a. Look for the maximum neighbour for the actual pixel.
				int maxDir = -1;
				sFloat maxVal = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);

				for (int d = 0; d < 4; d++)
					if ( maxVal < QVIMAGE_PIXEL(cornerResponseImage, col + xD[d], row + yD[d], 0) )
						{
						maxDir = d;
						maxVal = QVIMAGE_PIXEL(cornerResponseImage, col + xD[maxDir], row + yD[maxDir], 0);
						}
	
				// b. If found, join actual pixel to its group.
				if (maxDir > -1)
					disjointSet.unify(col, row, col + xD[maxDir], row + yD[maxDir]);
				}

	// 2. Create sets for all the sets of pixels with more than one element.
	QMap<int, QVVector> hotPoints;
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			{
			const sFloat pixelValue = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row, 0) >= threshold)
				{
				QVVector v(3);
				v[0] = pixelValue * col;
				v[1] = pixelValue * row;
				v[2] = pixelValue;
				const int setIndex = disjointSet.find(col, row);
				if (hotPoints.contains(setIndex))
					hotPoints[setIndex] += v;
				else
					hotPoints.insert(setIndex, v);
				}
			}

	// 3. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	foreach(int index, hotPoints.keys())
		{
		QVVector v = hotPoints[index];
		sortedPoints.insertMulti(v[2], QPointF(v[0]/v[2]+2, v[1]/v[2]+2));
		}

	QList<QPointF> result;
	foreach(sFloat key, sortedPoints.uniqueKeys())
		foreach (QPointF point, sortedPoints.values(key))
		result.append(point);

	return result;
	}

QList< QPair<sFloat, QPointF> > GetMaximalResponsePoints1bis(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	const int sizeMax = 2;
	const int cols = cornerResponseImage.getCols(), rows = cornerResponseImage.getRows();
	QVImage<uChar> binaryResponseImage(cols, rows);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);
	QVIMAGE_INIT_WRITE(uChar,binaryResponseImage);

	// 1. Get image with local maximums
	Set(binaryResponseImage,0);
	sFloat actual;
	for(int row = sizeMax; row < rows-sizeMax; row++)
		for(int col = sizeMax; col < cols-sizeMax; col++)
			{
			actual = QVIMAGE_PIXEL(cornerResponseImage, col, row,0);
			if (actual >= threshold)
				{
				QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = IPP_MAX_8U;
				for (int j = row-sizeMax; (j < row+sizeMax) && (QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) > 0); j++)
					for (int i = col-sizeMax; i < col+sizeMax; i++)
						if ( ((i != col) || (j != row)) && (actual <= QVIMAGE_PIXEL(cornerResponseImage, i, j, 0)) )
							{
							QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = 0;
							break;
							}
				}
			}

	// 2. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	for(uInt row = 0; row < binaryResponseImage.getRows(); row++)
		for(uInt col = 0; col < binaryResponseImage.getCols(); col++)
			if (QVIMAGE_PIXEL(binaryResponseImage, col, row,0))
				sortedPoints.insertMulti(QVIMAGE_PIXEL(cornerResponseImage, col, row,0), QPointF(col+2, row+2));

	QList< QPair<sFloat, QPointF> > result;
	foreach(const sFloat key, sortedPoints.uniqueKeys())
		foreach (const QPointF point, sortedPoints.values(key))
			result.append(QPair<sFloat, QPointF>(key,point));

	return result;
	}

QList< QPair<sFloat, QPointF> > GetMaximalResponsePoints3bis(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	static const int	xD[4] = { -1, +0, +1, +0 },
				yD[4] = { +0, -1, +0, +1 };

	const int rows = cornerResponseImage.getRows(), cols = cornerResponseImage.getCols();
	QVDisjointSet disjointSet(cornerResponseImage);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);

	// 1. Join in the disjoint set all the neighbour pixels
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row,0) >= threshold)
				{
				// a. Look for the maximum neighbour for the actual pixel.
				int maxDir = -1;
				sFloat maxVal = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);

				for (int d = 0; d < 4; d++)
					if ( maxVal < QVIMAGE_PIXEL(cornerResponseImage, col + xD[d], row + yD[d], 0) )
						{
						maxDir = d;
						maxVal = QVIMAGE_PIXEL(cornerResponseImage, col + xD[maxDir], row + yD[maxDir], 0);
						}
	
				// b. If found, join actual pixel to its group.
				if (maxDir > -1)
					disjointSet.unify(col, row, col + xD[maxDir], row + yD[maxDir]);
				}

	// 2. Create sets for all the sets of pixels with more than one element.
	QMap<int, QVVector> hotPoints;
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			{
			const sFloat pixelValue = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row, 0) >= threshold)
				{
				QVVector v(3);
				v[0] = pixelValue * col;
				v[1] = pixelValue * row;
				v[2] = pixelValue;
				const int setIndex = disjointSet.find(col, row);
				if (hotPoints.contains(setIndex))
					hotPoints[setIndex] += v;
				else
					hotPoints.insert(setIndex, v);
				}
			}

	// 3. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	foreach(int index, hotPoints.keys())
		{
		QVVector v = hotPoints[index];
		sortedPoints.insertMulti(v[2], QPointF(v[0]/v[2]+2, v[1]/v[2]+2));
		}

	QList< QPair<sFloat, QPointF> > result;
	foreach(const sFloat key, sortedPoints.uniqueKeys())
		foreach (const QPointF point, sortedPoints.values(key))
			result.append(QPair<sFloat, QPointF>(key,point));

	return result;
	}


