/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcpuplot.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef CPUPLOT_H
#define CPUPLOT_H

#include <qvgui/qvplot.h>
#include <qvutils/qvcpustat.h>


/// @class QVCpuPlot qvgui/qvcpuplot.h
/// @brief Class for plot graphs of worker's cpustat properties.
///
/// This class lets us plot graphs for QVStat worker's output properties. First, we need workers with QVStat output properties, like this:
///
///\code
///class MyWorker: public QVWorker
///	{
///	public:
///		MyWorker(QString name): QVWorker(name)
///			{
///			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
///			addProperty< QVImage<uChar,1> >("Output image", outputFlag);
///			addProperty<QVStat>("statistics", outputFlag);
///
///			...
///\endcode
///
/// And then, in the main function, we can create a QVCpuPlot and link it with that property, like this:
///
///\code
///...
///
///MyWorker myWorker("worker");
///QVCpuPlot cpuPlot("statistics plot");
///cpuPlot.linkProperty(myWorker, "statistics");
///
///...
///\endcode
///
///Showing us a graph window like this:
///
///\image html qvision_cpuplot_window.png
///
///If we link some workers QVStat properties to one QVCpuPlot, it plots all QVStat's time flags in link order, like this:
///
///\image html qvision_somecpuplots_window.png
///
///
///All workers have a QVStat output property, "cpu stats", that can be linked like this:
///
///\code
///...
///
///MyWorker myWorker("worker");
///QVCpuPlot cpuPlot("statistics plot");
///cpuPlot.linkProperty(myWorker);
///
///...
///\endcode
///
///And it's QVCpuPlot graph is shown when we puss the graph worker GUI button:
///\image html qvision_gui_graphbutton.png
///And plots the cpu time between each timeFlag() in the worker's iterate().
///
/// See also: @ref QVNumericPlot class @ref QVHistogramPlot class.
///
/// @ingroup qvgui

class QVCpuPlot : public QVPlot
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVCpuPlot with a given string identifier, showing properties and iteration's parameters.
		///
		/// @param name Identifier for the object to be created.
		/// @param decorations Indicate if the plot must shown the scale axis.
		/// @param autoShow Indicate if the plot are shown at the beginning or must be shown by the user.
		/// @param time True if the iterations are based on clock time (in second hundredths), and false if they are based on worker's iterations.
		/// @param step Set the base measure to plot's iterations ratio.
		/// @param parent Identifier for the parent widget.
		QVCpuPlot(const QString name = QString(), bool decorations = true, bool autoShow = true, bool time = true, int step = 100, QWidget *parent = 0);

		/// @brief Links properties from workers.
		/// This method safely links the "cpu stats" worker's property  in a thread safe manner.
		///
		/// See also QVPropertyContainer::LinkType.
		/// @param worker The source QVWorker.
		/// @param propertyName Identifying QString for the source property.
		void linkProperty(QVWorker &worker);

		/// @brief Links properties from workers.
		///
		/// This method safely links QVStat properties from workers in a thread safe manner.
		/// If the iterations are based on worker's iterations (time == false), the worker must be synchronised with previous linked workers.
		/// Only can be linked QVStat properties.
		///
		/// See also linkProperty(QVWorker &worker, const QString propertyName).
		/// @param worker The source QVWorker.
		void linkProperty(QVWorker &worker, const QString propertyName);

	private:
		/// @brief Returns a list with a int or double property curve's names.
		/// Return a list with the flag's names of the given cpustat property, one for each worker's time flag.
		///
		/// @param property The property name.
		QStringList getPropertyCurvNames(QString property) const;

		/// @brief Returns a list with a int or double property curve's value.
		/// Return a list with the time value of the property flags, one for each worker's time flag.
		///
		/// @param property The property name.
		QList<double> getPropertyCurvValues(QString property) const;

		/// @brief Returns a list with a int or double property curve's index.
		/// Return a list with the order with the flags appears in the worker. In order to a time flag that not appear in the firsts iterations can't be
		/// shown out of order.
		///
		/// @param property The property name.
		QList<int> getPropertyCurvOrders(QString property) const;
	};

#endif
