/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvhistogramplot.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <iostream>
#include <qvgui/qvhistogramplot.h>
#include <qwt_interval_data.h>
#include <qwt_scale_draw.h>
#include <qwt_scale_div.h>
#include <qwt_scale_engine.h>


QVHistogramPlot::QVHistogramPlot(const QString name, bool time, int step, double maxim, double minim, QWidget *parent):
	QVPlot(name, false, false, false, true, time, step, parent), histItem(), max(maxim), min(minim)
	{
	histItem.setColor(Qt::darkCyan);
	enableAxis(0,true);
	enableAxis(2,true);
	setAxisScaleDraw(QwtPlot::xBottom, new QwtScaleDraw());
	setAxisTitle(QwtPlot::xBottom, "Index");
	}

void QVHistogramPlot::linkProperty(QVWorker &worker, const QString propertyName)
	{
	int propType = worker.getPropertyType(propertyName);
	QList<double> auxList;
	int histType = QVariant::fromValue(auxList).userType();

	if (propType != histType)
		std::cerr << "Warning: a histogramplot only can be linked to a int or double property." << std::endl;
	else
		QVPlot::linkProperty(worker, propertyName);
	}


QStringList QVHistogramPlot::getPropertyCurvNames(QString property) const
	{
	QStringList names;
	QList<double> auxList = getPropertyValue<QList<double> >(property);
	for (int i = 0; i < auxList.size(); i++)
		names << QString("%1").arg(i);
	return names;
	}

QList<double> QVHistogramPlot::getPropertyCurvValues(QString property) const
	{
	return getPropertyValue<QList<double> >(property);
	}

QList<int> QVHistogramPlot::getPropertyCurvOrders(QString property) const
	{
	QList<int> order;
	QList<double> auxList = getPropertyValue<QList<double> >(property);
	for (int i = 1; i <= auxList.size(); i++)
		order << i;
	return order;
	}

void QVHistogramPlot::init()
	{
	if (initied)
		{
		std::cerr << "Warning: a plot can't be initied more than one time." << std::endl;
		return;
		}

	histItem.attach(this);

	readInputProperties();
	for(int i = 0; i < linkCont.size(); i++)
		for(int j = 0; j < linkCont[i].properties.size(); j++)
			{
			const QStringList curvNames = getPropertyCurvNames(linkCont[i].properties[j].name);

			for(int k = curvNames.size()-1; k >= 0; k--)
				{
				QwtPlotCurve * qwtpc;
				if (byTime) linkCont[i].properties[j].curves.prepend(Curve(curvNames.at(k), qwtpc, 1));
				else	    linkCont[i].properties[j].curves.prepend(Curve(curvNames.at(k), qwtpc, linkCont.size()+1));
				haveCurves = TRUE;
				}
			}

	if (byTime) startTimer(nStep * 10); // nStep * 10 ms

	initied = true;
	}

void QVHistogramPlot::insertNewFlags(int cont, int prop)
	{
	const QStringList curvNames = getPropertyCurvNames(linkCont[cont].properties[prop].name);
	if ( (linkCont.size() > cont) && (linkCont[cont].properties.size() > prop) && (curvNames.size() > linkCont[cont].properties[prop].curves.size()) )
		{
		const QList<int> curvOrders = getPropertyCurvOrders(linkCont[cont].properties[prop].name);
		for (int i = 0; i < curvOrders.size(); i++)
			if (curvOrders.at(i) > linkCont[cont].properties[prop].curves.size())
				{
				QwtPlotCurve * qwtpc;
				if (byTime) linkCont[cont].properties[prop].curves.insert(i, Curve(curvNames.at(i), qwtpc, 1));
				else	    linkCont[cont].properties[prop].curves.insert(i, Curve(curvNames.at(i), qwtpc, linkCont.size()+1));
				haveCurves = TRUE;
				}
		}
	}

void QVHistogramPlot::advancePlot()
	{
	if (!haveCurves)
		{
		std::cerr << "QVPlot internal error: early call to advancePlot." << std::endl;
		return;
		}

	// update data
	int numValues = 0;
	for(int i = 0; i < linkCont.size(); i++)
		for(int j = 0; j < linkCont[i].properties.size(); j++)
			for(int k = 0; k < linkCont[i].properties[j].curves.size(); k++)
				{
				updateValue(i, j, k); // and puss the new value
				numValues++;
				}

	QwtArray<QwtDoubleInterval> intervals(numValues);
	QwtArray<double> values(numValues);

	// curve update
	int index = 0;
	for(int i = 0; i < linkCont.size(); i++)
		for(int j = 0; j < linkCont[i].properties.size(); j++)
			for(int k = 0; k < linkCont[i].properties[j].curves.size(); k++)
				{
				values[index] = linkCont[i].properties[j].curves[k].history[0];
				intervals[index] = QwtDoubleInterval(double(index)-0.5, double(index+1)-0.5);
				index++;
				}

	// scroll bottom axis, scaled like an index
	int scaleShow = MAX(index/20, 1);
	setAxisScaleDiv(QwtPlot::xBottom, axisScaleEngine(QwtPlot::xBottom)->divideScale(-0.5, double(index)-0.5, 1, 1, scaleShow));

	histItem.setData(QwtIntervalData(intervals, values));

	// adjust left axis (using the visible values's maximum)
	if (max == 0)
		{
		for(int i = 0; i < linkCont.size(); i++)
			for(int j = 0; j < linkCont[i].properties.size(); j++)
				for(int k = 0; k < linkCont[i].properties[j].curves.size(); k++)
					if (max < linkCont[i].properties[j].curves[k].history[0]) max = linkCont[i].properties[j].curves[k].history[0];
		max = 1.1 *max;
		setAxisScale(QwtPlot::yLeft, min, max);
		max = 0;
		}
	else setAxisScale(QwtPlot::yLeft, min, max);

	// and replot
	replot();
	}

