/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimagecanvas.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGECANVAS_H
#define QVIMAGECANVAS_H

#include <qvgui/qvcanvas.h>
#include <qvcore/qvpropertycontainer.h>
#include <qvcore/qvworker.h>
#include <qvip/qvpolyline.h>

#include <QPointF>

/*!
@class QVImageCanvas qvimagecanvas.h qvgui/qvimagecanvas.h
@brief Viewer widget for @ref QVImage objects and other data types.

When created an object derived from this class, the interface opens a window that depicts the content of a property of type QVImage contained in a property holder (normally a @ref QVWorker object).

Usage is as follows: first create a QVImageCanvas object in the <i>main()</i> function of a program, then link a QVImage type property from a property holder:

\code
[...]

CannyWorker cannyWorker("Canny");

[...]

QVImageCanvas imageCanvas;
imageCanvas.linkProperty(cannyWorker,"Canny image");

[...]
\endcode

and latter call for the <i>exec()</i> function of the @ref QVApplication object of the program. This will automatically create a window like the following:

\image html qvimagecanvas_penguin.png

You can see it has a zoom number indicator (<i>z = 1</i>), horizontal and vertical rules (in pixels), and some buttons. These latter control zooming and moving around a zoomed area of the image mostly, and a detailed explanation of each one follows:

<table>
<tr><th><b>Zoom in button</b></th>
\image html qvimagecanvas_zoomin.png
<td>This button zooms the image in. Canvas window doesn't change size, but the area of the image displayed becomes smaller by a factor of 4 (width and height gets divided by 2). The label of the left bottom corner displays the text
\code
z=<zoom>
\endcode
where <i>&lt;zoom&gt;</i> is a number indicating the zoom factor that divides width and height in the actual zoom level.</td></tr>

<tr><th><b>Zoom out button</b></th>
\image html qvimagecanvas_zoomout.png
<td>This button divides zoom factor by 2, if it is equal or greater than 2. If the canvas window is bigger than the image at the final zoom, it is adjusted to the size of the latter.</td></tr>

<tr><th><b>Zoom restore button</b></th>
\image html qvimagecanvas_zoomrestore.png
<td>This button sets zoom factor by 1, and resizes canvas window to its original size.</td></tr>

<tr><th><b>Select zoom region button</b></th>
\image html qvimagecanvas_zoomregion.png
<td>This button lets you select a rectangle in the image to zoom, adjusting zoom factor and canvas window size to it.</td></tr>

<tr><th><b>Select polyline button</b></th>
\image html qvimagecanvas_polyline.png
<td>This button lets you select points in the image. That points list can be get as the "poly select" imageCanvas's property. You can explain this button by pushing it for a few seconds, and you can select a polyline representation, a point list representation, or generate a circular polyline, by a click and drag mode.</td></tr>

<tr><th><b>Select ROI button</b></th>
\image html qvimagecanvas_roi.png
<td>This button lets you select a rectangle in the image, that rectangle can be get as the "rect select" imageCanvas's property. And can be use to apply operations only to this rectangle.</td></tr>

<tr><th><b>Drag mode</b></th>
\image html qvimagecanvas_drag.png
<td>This button activates drag mode. With it you can move around the image displayed in the canvas window, if the zoom factor forces the canvas window to show only a sub-region of the image, by holding click and dragging over the depicted image.</td></tr>
</table>

In the following figure you can see a zoomed canvas window depicting a sub-region of the original image:

\image html qvimagecanvas_penguinpixels.png

Another interesting feature of image canvas is that at a zoom factor bigger or equal to 32, the canvas renders the gray-scale pixel value over every pixel if the image is gray-scale, or the three different values for each of the RGB channels over every pixel, if the image is RGB, as depicted below:

\image html qvimagecanvas_penguinpixelnumbers.png

@ingroup qvgui
*/

class QVImageCanvas: public QVCanvas, public QVPropertyContainer
	{
		Q_OBJECT
	public:
		QVImageCanvas(const QString name = QString (), QWidget *parent=0);
		void linkProperty(QVWorker &worker, const QString propertyName, QColor color = Qt::red, bool printTags = false);
		bool linkROI(QVWorker *worker, const QString propertyName);
		bool linkPolyline(QVWorker *worker, const QString propertyName);
		void setLowHigh(float low,float high);
		#ifndef DOXYGEN_IGNORE_THIS
		void viewer();
		#endif
		virtual void custom_viewer() { };

	public slots:
		void rectSelectedSlot(QRect rect);
		void polySelectedSlot(QPoint point, bool reset, TPolyMode mode);
		void circleSelectedSlot(QPoint center, float radius);

	private:
		float _low,_high;
		QPoint circleCenter;
		void draw(const QList<QPoint> &pointList, QColor color = Qt::red, bool printTags = false);
		void draw(const QList<QPointF> &pointList, QColor color = Qt::red, bool printTags = false);
		void draw(const QList< QList< QPointF > > &pointList, QColor color = Qt::red, bool printTags = false);
		void draw(const QVPolyline &contours, QColor color = Qt::red, bool printTags = false, TPolyMode mode = LINE);
		void draw(const QPoint &center, float radius, QColor color = Qt::red, bool printTags = false);
		void draw(const QList<QVPolyline> &contours, QColor color = Qt::red, bool printTags = false);
	};

Q_DECLARE_METATYPE(QList<QPoint>);
Q_DECLARE_METATYPE(QList<QPointF>);
Q_DECLARE_METATYPE(QList< QList< QPointF > >);
Q_DECLARE_METATYPE(QVPolyline);
Q_DECLARE_METATYPE(TPolyMode);

#endif
