/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvplot.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVPLOT_H
#define QVPLOT_H

#include <QTimer>
#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qvcore/qvworker.h>

#ifndef DOXYGEN_IGNORE_THIS
#define MAX_HISTORY 60 // seconds

/// @class QVPlot qvgui/qvplot.h
/// @brief Class for plot worker's graphs.
///
/// This virtual class plots graphs for worker's properties that have been linked previously. But it doesn't know what type of properties are linked,
/// or how to gets the properties values. Each plot class must restrict the type of linked properties, and implements the getPropertyCurvNames, getPropertyCurvValues
/// and getPropertyCurvOrders functions.
///
/// @ingroup qvgui

class QVPlot: public QwtPlot, public QVPropertyContainer
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVPlot with a given string identifier, showing properties and iteration's parameters.
		///
		/// @param name Identifier for the object to be created.
		/// @param decorations Indicate if the plot must shown the scale axis.
		/// @param havePie Indicate if the plot must shown a top-left pie graph.
		/// @param brush Indicate if the graph's data values will be count form the axis or from the previous value.
		/// @param autoShow Indicate if the plot are shown at the beginning or must be shown by the user.
		/// @param time True if the iterations are based on clock time (in second hundredths), and false if they are based on worker's iterations.
		/// @param step Set the base measure to plot's iterations ratio.
		/// @param parent Identifier for the parent widget.
		QVPlot(const QString name = QString(), bool decorations = true, bool havePie = false, bool brush = false, bool autoShow = true, bool time = true,
		       int step = 10, QWidget *parent = 0);

		/// @brief Links properties from workers.
		///
		/// This method safely links properties from workers in a thread safe manner.
		/// If the iterations are based on worker's iterations (time == false), the worker must be synchronized with previous linked workers.
		///
		/// See also QVPropertyContainer::LinkType.
		/// @param worker The source QVWorker.
		/// @param propertyName Identifying QString for the source property.
		virtual void linkProperty(QVWorker &worker, const QString propertyName);

	public slots:
		/// @brief Initialice the plot.
		/// A plot must be initialised after links all properties, and before update or shown the graphs.
		///
		virtual void init();

		/// @brief Update the worker properties value of this worker.
		///
		/// @param id The worker id whose properties value must be updated.
		/// @param iteration The current workers iteration.
		void update(uint id, int iteration);

		/// @brief Inform the plot if a linked worker has finish.
		/// In a plot based in worker's iterations, it inform that the plot must wait less worker's update to advance the graph.
		///
		/// @param status A worker status, it only matter if there are finished.
		void workerChange(QVWorker::TWorkerStatus status);
 
 	protected:
		/// @brief Returns a list with a property curve's names.
		/// A property can contain several values to be plot, each value must have a name.
		///
		/// @param property The property name.
		virtual QStringList getPropertyCurvNames(QString property) const = 0;

		/// @brief Returns a list with a property curve's values.
		/// A property can contain several values to be plot.
		///
		/// @param property The property name.
		virtual QList<double> getPropertyCurvValues(QString property) const = 0;

		/// @brief Returns a list with a property curve's index.
		/// A property can contain several values to be plot, each value must have a index that indicate the order to be plot. This value is useful
		/// in brush mode plots.
		///
		/// @param property The property name.
		virtual QList<int> getPropertyCurvOrders(QString property) const = 0;

			class Curve
				{
				public:
					Curve(QString n, QwtPlotCurve *qwtpc, int tempSize): name(n), plot(qwtpc)
						{
						for (int i = 0; i < MAX_HISTORY; i++) history[i] = 0;
						if (tempSize > 0)
							{
							temp = new double[tempSize];
							for (int i = 0; i < tempSize; i++) temp[i] = 0;
							}
						}
					QString name;
					QwtPlotCurve * plot;
					double history[MAX_HISTORY];
					double * temp;
				};
	
			class Property
				{
				public:
					Property(QString n): name(n) {}
					QString name;
					QList<Curve> curves;
				};
	
			class LinkedContainer
				{
				public:
					LinkedContainer(int ident): id(ident), iter(0), meanItems(0) {}
					uint id;
					int iter;
					QList<Property> properties;
					int meanItems;
				};

		QList<LinkedContainer> linkCont;
		const bool byTime;
		int iterationIndex;

		QColor nextColor();
		virtual void advancePlot();
		void timerEvent(QTimerEvent *e);
		virtual void insertNewFlags(int cont, int prop);
		void updateValue(const int cont, const int prop, const int curv);
		double getValue(const int cont, const int prop, const int curv) const;

		static QColor colors[];
		const bool decorations, hPie, doBrush;
		const int nStep;

			friend class PieMarker;
			class PieMarker: public QwtPlotItem
				{
				public:
					PieMarker(QVPlot *plot);
					virtual int rtti() const;
					virtual void draw(QPainter *p, const QwtScaleMap &, const QwtScaleMap &, const QRect &rect) const;
			
				private:
					QVPlot *qvplot;
				};

		bool initied;
		QTimer timer;
		PieMarker *pie;
		bool haveCurves;
		int activeWorkers;
		QList<QVPropertyContainer> pcl_Workers;
		double timeData[MAX_HISTORY];
		int dataCount;
		int usedColors;
	};

#endif
#endif
