/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvip.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvip/qvipp/qvipp.h>
#include <qvip/qvip.h>

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

void FilterLocalMax(const QVImage<sFloat> &src, QVImage<uChar> &dest, uInt colMaskSize, uInt rowMaskSize, sFloat threshold)
	{
	const int cols = src.getCols(), rows = src.getRows();
	Set(dest,0);
	sFloat actual;
	QVIMAGE_INIT_READ(sFloat,src);
	QVIMAGE_INIT_WRITE(uChar,dest);
	for(int row = rowMaskSize; row < rows-rowMaskSize; row++)
		for(int col = colMaskSize; col < cols-colMaskSize; col++)
			{
			actual = QVIMAGE_PIXEL(src, col, row,0);
			if (actual >= threshold)
				{
				QVIMAGE_PIXEL(dest, col, row, 0) = IPP_MAX_8U;
				for (int j = row-rowMaskSize; (j < row+rowMaskSize) && (QVIMAGE_PIXEL(dest, col, row, 0) > 0); j++)
					for (int i = col-colMaskSize; i < col+colMaskSize; i++)
						if ( ((i != col) || (j != row)) && (actual <= QVIMAGE_PIXEL(src, i, j, 0)) )
							{
							QVIMAGE_PIXEL(dest, col, row, 0) = 0;
							break;
							}
				}
			}
	}

void FilterHarrisCornerResponseImage(const QVImage<uChar> &image, QVImage<sFloat> &result)
	{
	const uInt rows = image.getRows(), cols = image.getCols();
	QVImage<uChar> buffer;
	MinEigenValGetBufferSize(image, buffer);
	QVImage<sFloat>	eigval(cols, rows);

	MinEigenVal(image, eigval, buffer);

	result = QVImage<sFloat>(cols, rows);

	FilterEqualizeHistogram(eigval, result);
	}

void FilterDoG(const QVImage<uChar> &image, QVImage<sFloat> &result)
	{
	const uInt rows = image.getRows(), cols = image.getCols();
	QVImage<uChar> gauss3x3(cols, rows), gauss5x5(cols, rows);
	
	FilterGauss(image, gauss3x3, 3);
	FilterGauss(image, gauss5x5, 5);

	result = QVImage<sFloat>(cols, rows);
	AbsDiff(gauss3x3, gauss5x5, result);	
	}

void SobelCornerResponseImage(const QVImage<sFloat> &image, QVImage<sFloat> &result)
	{
	std::cerr << "WARNING: SobelCornerResponseImage is deprecated. Use FilterHessianCornerResponseImage instead." << std::endl;
	FilterHessianCornerResponseImage(image, result);
	}

void FilterHessianCornerResponseImage(const QVImage<sFloat> &image, QVImage<sFloat> &result)
	{
	const uInt	cols = image.getCols(), rows = image.getRows();
	//const QVImage<sFloat> imageSFloat = image;

	QVImage<sFloat> Gx(cols, rows), Gy(cols, rows), Gxx(cols, rows),
			Gxy(cols, rows), Gyx(cols, rows), Gyy(cols, rows);

	FilterSobelHorizMask(image,Gx);
	FilterSobelVertMask(image,Gy);

	FilterSobelHorizMask(Gx,Gxx);
	FilterSobelVertMask(Gy,Gyy);

	FilterSobelHorizMask(Gy,Gyx);
	FilterSobelVertMask(Gx,Gxy);

	QVImage<sFloat> GxyGyx(cols, rows), GxxGyy(cols, rows);

	Mul(Gxy, Gyx, GxyGyx);
	Mul(Gxx, Gyy, GxxGyy);

	QVImage<sFloat> diffGxyGyx_GxxGyy(cols, rows);

	diffGxyGyx_GxxGyy.setAnchor(4,4);

	Sub(GxyGyx, GxxGyy, diffGxyGyx_GxxGyy);

	QVIMAGE_INIT_WRITE(sFloat,diffGxyGyx_GxxGyy);
	for (uInt col = 0; col < cols; col++)
		for(uInt row = 0; row < rows; row++)
			if (QVIMAGE_PIXEL(diffGxyGyx_GxxGyy, col, row, 0) <= 0)
				QVIMAGE_PIXEL(diffGxyGyx_GxxGyy, col, row, 0) = 0;

	result = QVImage<sFloat>(cols, rows);

	FilterEqualizeHistogram(diffGxyGyx_GxxGyy, result);
	}


int myFloodFill(QVImage<uChar> &image, uInt x, uInt y, uInt value, uInt minVal, uInt maxVal)
	{
	// Value should be inside range [minVal, maxVal]
	Q_ASSERT( (value <= minVal) || (value >= maxVal) );
	Q_ASSERT( minVal <= maxVal );

	// Coordinates should be inside the image.
	if ( (x >= image.getCols()) || (y >= image.getRows()))
		return 0;

	if ( (image(x,y) < minVal) || (image(x,y) > maxVal) )
		return 0;

	image(x,y) = value;

	int val = 1;
	val += myFloodFill(image, x-1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y+1, value, minVal, maxVal);

	val += myFloodFill(image, x-1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x-1, y+1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y+1, value, minVal, maxVal);

	return val;
	}

////////////////////////////

#include <qvip/qvimagefeatures/qvcomponenttree.h>

void pruneLowComponentTreeAux(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea, uInt node, uInt validThreshold)
	{
	Q_ASSERT(componentTree.area(node)[componentTree.firstThreshold(node)] != 0);
	Q_ASSERT(componentTree.area(node)[componentTree.lastThreshold(node)] != 0);
	Q_ASSERT(validThreshold >= componentTree.lastThreshold(node));

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Here we decide if this node should be directly pruned
	// or if there's any sub-node we should prune
	for (int threshold = componentTree.lastThreshold(node); threshold >= componentTree.firstThreshold(node) && !prune; threshold--)
		if (componentTree.area(node)[threshold] > 0)
			{
			if (componentTree.area(node)[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// We prune node, or get on with it's childrens
	if (prune)
		myFloodFill(image, componentTree.seedX(node), componentTree.seedY(node), lastValidThreshold, 0, lastValidThreshold-1);
	else
		for (uInt child = componentTree.firstChild(node); child != NULL_NODE; child = componentTree.nextSibling(child))
			pruneLowComponentTreeAux(image, componentTree, minArea, child, lastValidThreshold);
	}

void pruneHighComponentTreeAux(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea, uInt node, uInt validThreshold)
	{
	Q_ASSERT(componentTree.area(node)[componentTree.firstThreshold(node)] != 0);
	Q_ASSERT(componentTree.area(node)[componentTree.lastThreshold(node)] != 0);
	Q_ASSERT(validThreshold >= componentTree.lastThreshold(node));

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Here we decide if this node should be directly pruned
	// or if there's any sub-node we should prune
	for (int threshold = componentTree.lastThreshold(node); threshold >= componentTree.firstThreshold(node) && !prune; threshold--)
		if (componentTree.area(node)[threshold] > 0)
			{
			if (componentTree.area(node)[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// We prune node, or get on with it's childrens
	if (prune)
		myFloodFill(image, componentTree.seedX(node), componentTree.seedY(node), 255-lastValidThreshold, 255-lastValidThreshold+1, 255-0);
	else
		for (uInt child = componentTree.firstChild(node); child != NULL_NODE; child = componentTree.nextSibling(child))
			pruneHighComponentTreeAux(image, componentTree, minArea, child, lastValidThreshold);
	}

void FilterPruneComponentTreeSmallRegions(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea)
	{
	qDebug() << "pruneRegions()";
	if (componentTree.isInverseTree())
		{
		if(componentTree.area(componentTree.rootNode())[componentTree.lastThreshold(componentTree.rootNode())] > minArea)
			pruneHighComponentTreeAux(image, componentTree, minArea, componentTree.rootNode(), componentTree.lastThreshold(componentTree.rootNode()));
		}
	else	{
		if(componentTree.area(componentTree.rootNode())[componentTree.lastThreshold(componentTree.rootNode())] > minArea)
			pruneLowComponentTreeAux(image, componentTree, minArea, componentTree.rootNode(), componentTree.lastThreshold(componentTree.rootNode()));
		}

	qDebug() << "pruneRegions() <~ return";
	}

////////////////////////////

#define DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(TYPE, C)				\
void FilterEqualizeHistogram(const QVImage<TYPE,C> &image, QVImage<TYPE,C> &equalized)	\
	{										\
	uInt	rows = image.getRows(), cols = image.getCols();				\
	TYPE	maxVal, minVal;								\
											\
	Max(image,maxVal);								\
	Min(image,minVal);								\
											\
	QVImage<TYPE,C> temp(cols, rows), result(cols, rows);				\
	SubC(image, temp, minVal);							\
	MulC(temp, result, 255/(maxVal-minVal));					\
	equalized = result;								\
	}

DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(uChar,1);
DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(sFloat,1);



