/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvipp.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvip/qvipp/qvipp.h>
#include <qvmath/qvmatrix.h>

// Macro to call IPP functions
#define	IPP_DEBUG(function, ipp_function, ...)				\
	{								\
	IppStatus status = ipp_function(__VA_ARGS__);			\
	Q_ASSERT_X(	status == ippStsNoErr, #ipp_function,		\
			ippGetStatusString(status));			\
	Q_UNUSED(status);						\
	}								\

// Reads a pointer to the image, counting the ROI
#define PDATA_READ(Image)	\
	(&Image.getReadData()[Image.getROI().y()*Image.getStep()/Image.getTypeSize() + Image.getChannels()*Image.getROI().x()])
#define PDATA_WRITE(Image)	\
	(&Image.getWriteData()[Image.getROI().y()*Image.getStep()/Image.getTypeSize() + Image.getChannels()*Image.getROI().x()])

#define PDATA_READ_MARGIN(Image, margin)						\
	(&Image.getReadData()[								\
		(Image.getROI().y()+margin)*Image.getStep()/Image.getTypeSize()		\
		+ Image.getChannels()*(Image.getROI().x()+margin)]			\
		)

#define PDATA_WRITE_MARGIN(Image, margin)						\
	(&Image.getWriteData()[								\
		(Image.getROI().y()+margin)*Image.getStep()/Image.getTypeSize()		\
		+ Image.getChannels()*(Image.getROI().x()+margin)]			\
		)

// Get the size of the image
#define	IMAGE_ROISIZE(Image)	((IppiSize){ Image.getROI().width(), Image.getROI().height() })
#define	IMAGE_ROIRECT(Image)	((IppiRect){ Image.getROI().x(), Image.getROI().y(), Image.getROI().width(), Image.getROI().height() })
#define	IMAGE_SIZE(Image)	((IppiSize){ Image.getCols(), Image.getRows() })
#define	IPP_SIZE(X,Y)		( (IppiSize) { (X), (Y) })
#define	IPP_POINT(X,Y)		( (IppiPoint) { (X), (Y) })
#define	IPP_MASK_SIZE(X,Y)	(((X) == 1 && (Y) == 3)? ippMskSize1x3:		\
 				((X) == 1 && (Y) == 5)? ippMskSize1x5:		\
				((X) == 3 && (Y) == 1)? ippMskSize3x1:		\
 				((X) == 5 && (Y) == 1)? ippMskSize5x1:		\
 				((X) == 3 && (Y) == 3)? ippMskSize3x3:		\
				/*((X) == 5 && (Y) == 5)?*/ ippMskSize5x5)

#define	CHECK_COMPATIBLE_IMAGES(NAME, Image1, Image2)										\
	Q_ASSERT_X(Image1.getROI().size() == Image2.getROI().size(), "" # NAME, "different ROI sizes in source images");	\

/////////////////////////////////////////////////////////////
// START OF MACROS
/////////////////////////////////////////////////////////////

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(NAME, TYPE, C, SUBFIX)						\
void NAME(const QVImage<TYPE, C> &src, QVImage<TYPE, C> &dest, QPoint destROIOffset)				\
	{													\
	qDebug() << "" # NAME << "()";										\
	qDebug() << "" # NAME << ": calling ippi" # NAME # SUBFIX;						\
	dest.setROI(dest.getAnchor().x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());	\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	qDebug() << "" # NAME # SUBFIX << "() <~ return";							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, uShort, 1, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, uShort, 3, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sShort, 3, _16s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sInt, 1, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sInt, 3, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Copy, sFloat, 3, _32f_C3R);

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSharpen, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSharpen, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelHoriz, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelHoriz, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelHoriz, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelVert, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelVert, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(FilterSobelVert, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Not, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Not, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Abs, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Abs, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Ln, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Ln, sFloat, 3, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE(Sqr, sFloat, 1, _32f_C1R);


/// TODO: check that all input images have the same dimensions and source roi sizes, and same step
#define DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(NAME, TYPE, SUBFIX)								\
void NAME(const QVImage<TYPE, 1> &src1, const QVImage<TYPE, 1> &src2, const QVImage<TYPE, 1> &src3, QVImage<TYPE, 3> &dest, QPoint destROIOffset)	\
	{																		\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src1.getROI().width(), src1.getROI().height());			\
	const TYPE	* pSrc[3] = { PDATA_READ(src1), PDATA_READ(src2), PDATA_READ(src3) };					\
																\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,										\
				pSrc, src1.getStep(),										\
				PDATA_WRITE(dest), dest.getStep(),								\
				IMAGE_ROISIZE(dest));										\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(Copy, uChar, _8u_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(Copy, uShort, _16u_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(Copy, sShort, _16s_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(Copy, sInt, _32s_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C1_QVIMAGE_C3(Copy, sFloat, _32f_P3C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(NAME, TYPE, SUBFIX)						\
void NAME(const QVImage<TYPE, 1> &src, QVImage<TYPE, 3> &dest, QPoint destROIOffset)				\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
	const TYPE	* pSrc[3] = { PDATA_READ(src), PDATA_READ(src), PDATA_READ(src) };			\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				pSrc, src.getStep(),								\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(Copy, uChar, _8u_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(Copy, uShort, _16u_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(Copy, sShort, _16s_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(Copy, sInt, _32s_P3C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_C1_QVIMAGE_C3(Copy, sFloat, _32f_P3C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(NAME, TYPE1, TYPE2, C, SUBFIX)				\
void NAME(const QVImage<TYPE1, C> &src, QVImage<TYPE2, C> &dest, QPoint destROIOffset)				\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, uShort, 1, _8u16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sShort, 1, _8u16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sInt, 1, _8u32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sFloat, 1, _8u32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, uShort, 3, _8u16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sShort, 3, _8u16s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sInt, 3, _8u32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uChar, sFloat, 3, _8u32f_C3R);

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, uChar, 1, _16u8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, sInt, 1, _16u32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, sFloat, 1, _16u32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, uChar, 3, _16u8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, sInt, 3, _16u32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, uShort, sFloat, 3, _16u32f_C3R);

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, uChar, 1, _16s8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, sInt, 1, _16s32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, sFloat, 1, _16s32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, uChar, 3, _16s8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, sInt, 3, _16s32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sShort, sFloat, 3, _16s32f_C3R);

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sInt, uChar, 1, _32s8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2(Convert, sInt, uChar, 3, _32s8u_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_PLUSVALUE(NAME, TYPE1, TYPE2, C, SUBFIX)			\
void NAME(const QVImage<TYPE1, C> &src, QVImage<TYPE2, C> &dest, TYPE2 val, QPoint destROIOffset)		\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest), val);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_PLUSVALUE(Integral, uChar, sInt, 1, _8u32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_PLUSVALUE(Integral, uChar, sFloat, 1,  _8u32f_C1R);


#define DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(NAME, TYPE1, TYPE2, C, SUBFIX)			\
void NAME(const QVImage<TYPE1, C> &src, QVImage<TYPE2, C> &dest, IppRoundMode roundMode, QPoint destROIOffset)	\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest),roundMode);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, uChar, 1, _32f8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, uShort, 1, _32f16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, sShort, 1, _32f16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, uChar, 3, _32f8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, uShort, 3, _32f16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_T1_QVIMAGE_T2_IPPROUNDMODE(Convert, sFloat, sShort, 3, _32f16s_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(NAME, TYPE, SUBFIX)						\
void NAME(const QVImage<TYPE, 3> &src, QVImage<TYPE, 1> &dest, QPoint destROIOffset)				\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(RGBToGray, uChar, _8u_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(RGBToGray, uShort, _16u_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(RGBToGray, sShort, _16s_C3C1R);
//DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(RGBToGray, sInt, _32s_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1(RGBToGray, sFloat, _32f_C3C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(NAME, TYPE, SUBFIX)					\
void NAME(const QVImage<TYPE, 3> &src, QVImage<TYPE, 1> &dest, uChar channel, QPoint destROIOffset)		\
	{													\
	qDebug() << "" # NAME << "()";										\
	qDebug() << "" # NAME << ": calling ippi" # NAME # SUBFIX;						\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src) + channel, src.getStep(),					\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	qDebug() << "" # NAME # SUBFIX << "() <~ return";							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(Copy, uChar, _8u_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(Copy, uShort, _16u_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(Copy, sShort, _16s_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(Copy, sInt, _32s_C3C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_C3_QVIMAGE_C1_BIS(Copy, sFloat, _32f_C3C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(NAME, TYPE, C, SUBFIX)						\
void NAME(const QVImage<TYPE,C> &src1, const QVImage<TYPE,C> &src2, QVImage<uChar> &dest, IppCmpOp ippCmpOp, QPoint destROIOffset)	\
	{																\
	CHECK_COMPATIBLE_IMAGES(NAME, src1, src2);									\
															\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src1.getROI().width(), src1.getROI().height());		\
															\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,									\
				PDATA_READ(src1), src1.getStep(),							\
				PDATA_READ(src2), src2.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),							\
				IMAGE_ROISIZE(dest),									\
				ippCmpOp);										\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, uShort, 1, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, uShort, 3, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, sShort, 3, _16s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_UCHAR_C1_IPPCMPOP(Compare, sFloat, 3, _32f_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP(NAME, TYPE, SUBFIX)				\
void NAME(const QVImage<TYPE> &src, QVImage<uChar> &dest, TYPE constant, IppCmpOp ippCmpOp, QPoint destROIOffset)	\
	{														\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				constant,									\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest),								\
				ippCmpOp);									\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP(CompareC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP(CompareC, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP_2(NAME, TYPE, SUBFIX)				\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, TYPE constant, IppCmpOp ippCmpOp, QPoint destROIOffset)	\
	{														\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,	ippi ## NAME ## SUBFIX,									\
			PDATA_READ(src), src.getStep(),								\
			PDATA_WRITE(dest), dest.getStep(),							\
			IMAGE_ROISIZE(dest),									\
			constant, ippCmpOp);									\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP_2(Threshold, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP_2(Threshold, sShort, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_IPPCMPOP_2(Threshold, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_IPPCMPOP(NAME, TYPE, SUBFIX)		\
void NAME(QVImage<TYPE> &src_dest, TYPE constant, IppCmpOp ippCmpOp)			\
	{										\
	IPP_DEBUG(NAME,	ippi ## NAME ## SUBFIX,									\
			PDATA_WRITE(src_dest), src_dest.getStep(),						\
			IMAGE_ROISIZE(src_dest),								\
			constant, ippCmpOp);									\
	}
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_IPPCMPOP(Threshold, uChar, _8u_C1IR);
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_IPPCMPOP(Threshold, sShort, _16s_C1IR);
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_IPPCMPOP(Threshold, sFloat, _32f_C1IR);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_IPPCMPOP(NAME, TYPE, SUBFIX)						\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, TYPE constant1, TYPE constant2, IppCmpOp ippCmpOp, QPoint destROIOffset)	\
	{																\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,	ippi ## NAME ## SUBFIX,									\
			PDATA_READ(src), src.getStep(),								\
			PDATA_WRITE(dest), dest.getStep(),							\
			IMAGE_ROISIZE(dest),									\
			constant1, constant2, ippCmpOp);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_IPPCMPOP(Threshold_Val, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_IPPCMPOP(Threshold_Val, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_SCALAR_IPPCMPOP(NAME, TYPE, SUBFIX)	\
void NAME(QVImage<TYPE> &src_dest, TYPE constant1, TYPE constant2, IppCmpOp ippCmpOp)	\
	{										\
	IPP_DEBUG(NAME,	ippi ## NAME ## SUBFIX,						\
			PDATA_WRITE(src_dest), src_dest.getStep(),			\
			IMAGE_ROISIZE(src_dest),					\
			constant1, constant2, ippCmpOp);				\
	}
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_SCALAR_IPPCMPOP(Threshold_Val, uChar, _8u_C1IR);
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_SCALAR_IPPCMPOP(Threshold_Val, sShort, _16s_C1IR);
DEFINE_IPP_FUNCTION_INPLACE_QVIMAGE_SCALAR_SCALAR_IPPCMPOP(Threshold_Val, sFloat, _32f_C1IR);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_SCALAR_SCALAR(NAME, TYPE, SUBFIX)						\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, TYPE constant1, TYPE constant2, TYPE constant3, TYPE constant4, QPoint destROIOffset)	\
	{																	\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,	ippi ## NAME ## SUBFIX,									\
			PDATA_READ(src), src.getStep(),								\
			PDATA_WRITE(dest), dest.getStep(),							\
			IMAGE_ROISIZE(dest),									\
			constant1, constant2, constant3, constant4);						\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_SCALAR_SCALAR(Threshold_LTValGTVal, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_UCHAR_C1_SCALAR_SCALAR_SCALAR_SCALAR(Threshold_LTValGTVal, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(NAME, TYPE, SUBFIX)							\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, uInt colMaskRadius, uInt rowMaskRadius, QPoint destROIOffset)	\
	{														\
	dest.setROI(destROIOffset.x(), destROIOffset.y(),							\
		src.getROI().width()-colMaskRadius+1, src.getROI().height()-rowMaskRadius+1);			\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),							\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest),								\
				IPP_SIZE(colMaskRadius, rowMaskRadius),						\
				IPP_POINT(0,0));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMax,		uChar,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMax,		uShort,	_16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMax,		sFloat,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMin,		uChar,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMin,		uShort,	_16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMin,		sFloat,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterBox,		uChar,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterBox,		uShort,	_16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterBox,		sFloat,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMedian,	uChar,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_INT_INT(FilterMedian,	uShort,	_16u_C1R);

// MASK SIZED
#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(NAME, TYPE, C, SUBFIX)						\
void NAME(const QVImage<TYPE, C> &src, QVImage<TYPE, C> &dest, uInt maskSize, QPoint destROIOffset)			\
	{														\
	Q_ASSERT_X(	(maskSize == 3) || (maskSize == 5),								\
			"" #NAME, "Mask size incorrect. Allowed sizes: 3 and 5, for 3x3 and 5x5 mask sizes");		\
															\
	Q_ASSERT_X(src.getROI().width()-maskSize+1 > 0,"" #NAME, "dest ROI size < 1");					\
	Q_ASSERT_X(src.getROI().height()-maskSize+1 > 0,"" #NAME, "dest ROI size < 1");					\
															\
	dest.setROI(destROIOffset.x(), destROIOffset.y(),								\
		src.getROI().width()-maskSize+1, src.getROI().height()-maskSize+1);					\
															\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,									\
				PDATA_READ_MARGIN(src,(maskSize -1)/2), src.getStep(),					\
				PDATA_WRITE(dest), dest.getStep(),							\
				IMAGE_ROISIZE(dest), IPP_MASK_SIZE(maskSize, maskSize)					\
				);											\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterGauss,		uChar,	1,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterGauss,		sFloat,	1,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterGauss,		uChar,	3,	_8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterGauss,		sFloat,	3,	_32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterHipass,		uChar,	1,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterHipass,		sFloat,	1,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterLowpass,		uChar,	1,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterLowpass,		sFloat,	1,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterSobelHorizMask,	sFloat,	1,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterSobelVertMask,	sFloat,	1,	_32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterLaplace,		uChar,	1,	_8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterLaplace,		sShort,	1,	_16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_IPPMSKSIZE(FilterLaplace,		sFloat,	1,	_32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_BUFFER_IPPMSKSIZE(NAME, TYPE1, TYPE2, SUBFIX)				\
void NAME(const QVImage<TYPE1> &src, QVImage<TYPE2> &dest, QVImage<TYPE1> &buffer, uInt maskSize, QPoint destROIOffset)	\
	{														\
	Q_ASSERT_X(	(maskSize == 3) || (maskSize == 5),								\
			"" #NAME, "Mask size incorrect. Allowed sizes: 3 and 5, for 3x3 and 5x5 mask sizes");		\
															\
	Q_ASSERT_X(src.getROI().width()-maskSize+1 > 0,"" #NAME, "dest ROI size < 1");					\
	Q_ASSERT_X(src.getROI().height()-maskSize+1 > 0,"" #NAME, "dest ROI size < 1");					\
															\
	dest.setROI(destROIOffset.x(), destROIOffset.y(),								\
		src.getROI().width()-maskSize+1, src.getROI().height()-maskSize+1);					\
															\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,									\
				PDATA_READ_MARGIN(src,(maskSize -1)/2), src.getStep(),					\
				PDATA_WRITE(dest), dest.getStep(),							\
				IMAGE_ROISIZE(dest), IPP_MASK_SIZE(maskSize, maskSize),					\
				ippBorderRepl, 0, buffer.getWriteData());						\
	}
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_BUFFER_IPPMSKSIZE(FilterSobelNegVertBorder, uChar, sShort, _8u16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_BUFFER_IPPMSKSIZE(FilterSobelHorizBorder, uChar, sShort, _8u16s_C1R);

#define DEFINE_IPP_FUNCTION_QVGENERICIMAGE_BUFFER_IPPMSKSIZE(NAME, TYPE, SUBFIX)	\
void NAME(const QVGenericImage &src, QVImage<TYPE> &buffer, uInt maskSize)		\
	{										\
	Q_ASSERT_X(	(maskSize == 3) || (maskSize == 5),								\
			"" #NAME, "Mask size incorrect. Allowed sizes: 3 and 5, for 3x3 and 5x5 mask sizes");		\
															\
	int bufferSize;													\
															\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX, IMAGE_ROISIZE(src),						\
				IPP_MASK_SIZE(maskSize, maskSize), &bufferSize);					\
	buffer = QVImage<uChar>(bufferSize,1);										\
	}

DEFINE_IPP_FUNCTION_QVGENERICIMAGE_BUFFER_IPPMSKSIZE(FilterSobelNegVertGetBufferSize, uChar, _8u16s_C1R);
DEFINE_IPP_FUNCTION_QVGENERICIMAGE_BUFFER_IPPMSKSIZE(FilterSobelHorizGetBufferSize, uChar, _8u16s_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(NAME, TYPE, C, SUBFIX)	\
void NAME(QVImage<TYPE, C> &img, TYPE constant)				\
	{								\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,	constant,	\
				PDATA_WRITE(img), img.getStep(),	\
				IMAGE_ROISIZE(img));			\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(Set, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(Set, uShort, 1, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(Set, sShort, 1, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(Set, sInt, 1, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR(Set, sFloat, 1, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(NAME, TYPE, SUBFIX)						\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, TYPE constant, QPoint destROIOffset)			\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),	constant,					\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(OrC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(AndC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(XorC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(OrC, uShort, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(AndC, uShort, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(XorC, uShort, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(OrC, sInt, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(AndC, sInt, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(XorC, sInt, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(LShiftC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(RShiftC, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(LShiftC, uShort, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(RShiftC, uShort, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(LShiftC, sInt, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(RShiftC, sInt, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(MulC, sFloat, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(AddC, sFloat, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(SubC, sFloat, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR(DivC, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(NAME, TYPE, SUBFIX)					\
void NAME(const QVImage<TYPE, 3> &src, QVImage<TYPE, 3> &dest, TYPE constant, QPoint destROIOffset)		\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
	TYPE value[3] = {constant, constant, constant};								\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),	value,						\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(OrC, uChar, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(AndC, uChar, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(XorC, uChar, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(OrC, uShort, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(AndC, uShort, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(XorC, uShort, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(OrC, sInt, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(AndC, sInt, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(XorC, sInt, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(MulC, sFloat, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(AddC, sFloat, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(SubC, sFloat, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3(DivC, sFloat, _32f_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(NAME, TYPE, SUBFIX)					\
void NAME(const QVImage<TYPE, 3> &src, QVImage<TYPE, 3> &dest, TYPE constant, QPoint destROIOffset)		\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
	Ipp32u value[3] = {constant, constant, constant};							\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),	value,						\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(LShiftC, uChar, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(RShiftC, uChar, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(LShiftC, uShort, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(RShiftC, uShort, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(LShiftC, sInt, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_C3_SHIFT(RShiftC, sInt, _32s_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(NAME, TYPE1, TYPE2, SUBFIX)	\
void NAME(const QVImage<TYPE1> &src, TYPE2 & value)				\
	{									\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,				\
				PDATA_READ(src), src.getStep(),			\
				IMAGE_ROISIZE(src), &value);			\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Sum, uChar, sDouble, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Sum, sShort, sDouble, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Mean, uChar, sDouble, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Mean, sShort, sDouble, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Max, uChar, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Max, sShort, sShort, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Max, sFloat, sFloat, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Min, uChar, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Min, sShort, sShort, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_2(Min, sFloat, sFloat, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(NAME, TYPE, C, SUBFIX)		\
void NAME(QVImage<TYPE, C> &img, TYPE red, TYPE green, TYPE blue)			\
	{										\
	TYPE constant[3] = { red, green, blue };				\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,	constant,		\
				PDATA_WRITE(img), img.getStep(),		\
				IMAGE_ROISIZE(img));				\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(Set, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(Set, uShort, 3, _16u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(Set, sShort, 3, _16s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(Set, sInt, 3, _32s_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR(Set, sFloat, 3, _32f_C3R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(NAME, TYPE1, TYPE2, SUBFIX)	\
void NAME(const QVImage<TYPE1,3> &src, TYPE2 & valueR, TYPE2 & valueG, TYPE2 & valueB)	\
	{										\
	TYPE2	values[3];								\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,					\
				PDATA_READ(src), src.getStep(),				\
				IMAGE_ROISIZE(src), values);				\
											\
	valueR = values[0]; valueG = values[1]; valueB = values[2];			\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Sum, uChar, sDouble, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Sum, sShort, sDouble, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Mean, uChar, sDouble, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Mean, sShort, sDouble, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Max, uChar, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Max, sShort, sShort, _16s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Min, uChar, uChar, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_SCALAR_SCALAR_SCALAR_2(Min, sShort, sShort, _16s_C1R);

#include <iostream>
#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(NAME, TYPE, C, SUBFIX)						\
void NAME(const QVImage<TYPE,C> &src1, const QVImage<TYPE,C> &src2, QVImage<TYPE,C> &dest, int scale, QPoint destROIOffset)	\
	{															\
	CHECK_COMPATIBLE_IMAGES(NAME, src1, src2);								\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src1.getROI().width(), src1.getROI().height());	\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src1), src1.getStep(),						\
				PDATA_READ(src2), src2.getStep(), 						\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest), scale);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, uChar, 1, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, uChar, 1, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, uChar, 1, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, uChar, 1, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, uChar, 3, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, uChar, 3, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, uChar, 3, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, uChar, 3, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, uShort, 1, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, uShort, 1, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, uShort, 1, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, uShort, 1, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, uShort, 3, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, uShort, 3, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, uShort, 3, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, uShort, 3, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, sShort, 1, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, sShort, 1, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, sShort, 1, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, sShort, 1, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Add, sShort, 3, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Sub, sShort, 3, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Mul, sShort, 3, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR(Div, sShort, 3, _16s_C3RSfs);


#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(NAME, TYPE, C, SUBFIX)						\
void NAME(const QVImage<TYPE,C> &src1, const QVImage<TYPE,C> &src2, QVImage<TYPE,C> &dest, QPoint destROIOffset)	\
	{														\
	CHECK_COMPATIBLE_IMAGES(NAME, src1, src2);								\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src1.getROI().width(), src1.getROI().height());	\
														\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src1), src1.getStep(),						\
				PDATA_READ(src2), src2.getStep(), 						\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest));								\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Add, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Sub, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Mul, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Div, sFloat, 1, _32f_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Add, sFloat, 3, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Sub, sFloat, 3, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Mul, sFloat, 3, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(Div, sFloat, 3, _32f_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(MulScale, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(MulScale, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(AbsDiff, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE(AbsDiff, sFloat, 1, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(NAME, TYPE, SUBFIX)					\
void NAME(const QVImage<TYPE> &src, QVImage<TYPE> &dest, TYPE constant, int scale, QPoint destROIOffset)	\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),	constant,					\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest),scale);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, uChar, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, uChar, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, uChar, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, uChar, _8u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, uShort, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, uShort, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, uShort, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, uShort, _16u_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, sShort, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, sShort, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, sShort, _16s_C1RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, sShort, _16s_C1RSfs);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(NAME, TYPE, SUBFIX)				\
void NAME(const QVImage<TYPE, 3> &src, QVImage<TYPE, 3> &dest, TYPE constant, int scale, QPoint destROIOffset)	\
	{													\
	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());		\
	TYPE value[3] = {constant, constant, constant};								\
														\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,								\
				PDATA_READ(src), src.getStep(),	value,						\
				PDATA_WRITE(dest), dest.getStep(),						\
				IMAGE_ROISIZE(dest),scale);							\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, uChar, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, uChar, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, uChar, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, uChar, _8u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, uShort, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, uShort, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, uShort, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, uShort, _16u_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(MulC, sShort, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(AddC, sShort, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(SubC, sShort, _16s_C3RSfs);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_QVIMAGE_SCALAR_SCALE(DivC, sShort, _16s_C3RSfs);


#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_DOUBLE_DOUBLE_INTERPOLATION(NAME, TYPE, PLANES, SUBFIX)	\
void NAME(const QVImage<TYPE, PLANES> &src, QVImage<TYPE, PLANES> &dest, int interpolation)		\
	{												\
	double	factorX = (double) dest.getROI().width() / (double) src.getROI().width(),		\
		factorY = (double) dest.getROI().height() / (double) src.getROI().height();		\
													\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,							\
				src.getReadData(), IMAGE_SIZE(src), src.getStep(), IMAGE_ROIRECT(src),	\
				PDATA_WRITE(dest), dest.getStep(), IMAGE_ROISIZE(dest),			\
				factorX, factorY, interpolation);					\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_DOUBLE_DOUBLE_INTERPOLATION(Resize, uChar, 1, _8u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_DOUBLE_DOUBLE_INTERPOLATION(Resize, uChar, 3, _8u_C3R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_DOUBLE_DOUBLE_INTERPOLATION(Resize, uShort, 1, _16u_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_DOUBLE_DOUBLE_INTERPOLATION(Resize, sFloat, 1, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_BUFFER_KERNELSIZE_SFLOAT_BORDERTYPE_BORDERVALUE(NAME, TYPE, C, SUBFIX)										\
void NAME(const QVImage<TYPE, C> &src, QVImage<TYPE, C> &dest, QVImage<uChar> &buffer, sInt kernelSize, sFloat sfloat, IppiBorderType borderType, sFloat borderValue, QPoint destROIOffset)	\
	{																							\
	dest.setROI(destROIOffset.x(), destROIOffset.y(),					\
		src.getROI().width()-kernelSize+1, src.getROI().height()-kernelSize+1);		\
												\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,						\
				PDATA_READ(src), src.getStep(),					\
				PDATA_WRITE(dest), dest.getStep(),				\
				IMAGE_ROISIZE(dest),						\
				kernelSize, sfloat, borderType, borderValue,			\
				buffer.getWriteData());						\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_QVIMAGE_BUFFER_KERNELSIZE_SFLOAT_BORDERTYPE_BORDERVALUE(FilterGaussBorder, sFloat, 1, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_THRESHOLD_PIPPIPOINT_UINT_PUINT_IPPINORM_BORDER(NAME, TYPE, C, SUBFIX)					\
void NAME(const QVImage<TYPE, C> &img, QVImage<uChar> &buffer, QList<QPoint> &points, TYPE threshold, IppiNorm norm, uInt border, uInt maxPeakCount)	\
	{																		\
	int pPeakCount;									\
	IppiPoint *pPeak = new IppiPoint[maxPeakCount]();				\
											\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,					\
				PDATA_READ(img), img.getStep(), IMAGE_ROISIZE(img),	\
				threshold, pPeak, maxPeakCount, &pPeakCount,		\
				norm, border, buffer.getWriteData());			\
											\
	/*const uInt max = (pPeakCount < maxPeakCount)?pPeakCount:maxPeakCount-1;*/	\
	for (int n = 0; n < maxPeakCount; n++)						\
		points.append(QPoint(pPeak[n].x, pPeak[n].y));				\
											\
	delete pPeak;									\
	}


DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_THRESHOLD_PIPPIPOINT_UINT_PUINT_IPPINORM_BORDER(FindPeaks3x3, sInt, 1, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_THRESHOLD_PIPPIPOINT_UINT_PUINT_IPPINORM_BORDER(FindPeaks3x3, sFloat, 1, _32f_C1R);

#define DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_ROIWIDTH(NAME, TYPE, C, SUBFIX)			\
void NAME(const QVImage<TYPE,C> &image, QVImage<uChar> &buffer, QPoint destROIOffset)		\
	{											\
	int pBufferSize;									\
 														\
	IPP_DEBUG(Peaks3x3GetBufferSize,	ippi ## NAME ## SUBFIX, image.getROI().width(), &pBufferSize);	\
	buffer = QVImage<uChar>(pBufferSize,1);									\
	}

DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_ROIWIDTH(FindPeaks3x3GetBufferSize, sInt, 1, _32s_C1R);
DEFINE_IPP_FUNCTION_QVIMAGE_BUFFER_ROIWIDTH(FindPeaks3x3GetBufferSize, sFloat, 1, _32f_C1R);

////////////////////////////////////////////////////////////////////////////////////////////////////
// END OF MACROS
////////////////////////////////////////////////////////////////////////////////////////////////////

void FilterGaussGetBufferSize(const QVGenericImage &image, QVImage<uChar> &buffer, uInt kernelSize)
	{
	int pBufferSize;
	IPP_DEBUG(FilterGaussGetBufferSize,	ippiFilterGaussGetBufferSize_32f_C1R, IMAGE_ROISIZE(image), kernelSize, &pBufferSize);
	buffer = QVImage<uChar>(pBufferSize,1);
	}

void MinEigenValGetBufferSize(const QVGenericImage &image, QVImage<uChar> &buffer, uInt apertureSize, uInt avgWindow)
	{
	int pBufferSize;
	IPP_DEBUG(MinEigenVal,	ippiMinEigenValGetBufferSize_8u32f_C1R, IMAGE_ROISIZE(image), apertureSize, avgWindow, &pBufferSize);
	buffer = QVImage<uChar>(pBufferSize,1);
	}

void	CannyGetSize(const QVGenericImage &src, QVImage<uChar>	&buffer)
	{
	int cannyBufferSize;
	IPP_DEBUG(CannyGetSize, ippiCannyGetSize, IMAGE_ROISIZE(src),&cannyBufferSize);
	buffer = QVImage<uChar>(cannyBufferSize,1);
	}

void Canny(QVImage<sShort> &derivadaX, QVImage<sShort> &derivadaY,  
		QVImage<uChar> &destino, QVImage<uChar> &buffer, double low, double high, QPoint destROIOffset)
	{
	CHECK_COMPATIBLE_IMAGES(NAME,derivadaX, derivadaY);
	destino.setROI(destROIOffset.x(), destROIOffset.y(), derivadaX.getROI().width(), derivadaX.getROI().height());

	IPP_DEBUG(Canny,	ippiCanny_16s8u_C1R,
				PDATA_WRITE(derivadaX), derivadaX.getStep(),
				PDATA_WRITE(derivadaY), derivadaY.getStep(),
				PDATA_WRITE(destino), destino.getStep(),
				IMAGE_ROISIZE(derivadaX), low, high,
				buffer.getWriteData() );
	}

void Canny(QVImage<sFloat> &derivadaX, QVImage<sFloat> &derivadaY,  
		QVImage<uChar> &destino, QVImage<uChar> &buffer, double low, double high, QPoint destROIOffset)
	{
	CHECK_COMPATIBLE_IMAGES(NAME,derivadaX, derivadaY);
	destino.setROI(destROIOffset.x(), destROIOffset.y(), derivadaX.getROI().width(), derivadaX.getROI().height());

	IPP_DEBUG(Canny,	ippiCanny_32f8u_C1R,
				PDATA_WRITE(derivadaX), derivadaX.getStep(),
				PDATA_WRITE(derivadaY), derivadaY.getStep(),
				PDATA_WRITE(destino), destino.getStep(),
				IMAGE_ROISIZE(derivadaX), low, high,
				buffer.getWriteData() );
	}

void MinEigenVal(const QVImage<uChar> &src, QVImage<sFloat> &mineigval, QVImage<uChar> &buffer,
	uInt apertureSize, uInt avgWindow, IppiKernelType kernel, QPoint destROIOffset)
	{
	Q_ASSERT_X(	( (kernel == ippKernelSobel) && (apertureSize == 3 || apertureSize == 5) ) ||
			( (kernel == ippKernelSobel) && (apertureSize == 3) ),
			"MinEigenVal", "apertureSize value incorrect");

	mineigval.setROI(mineigval.getAnchor().x(), mineigval.getAnchor().y(), src.getROI().width(), src.getROI().height());

	IPP_DEBUG(MinEigenVal, ippiMinEigenVal_8u32f_C1R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(mineigval), mineigval.getStep(),
				IMAGE_ROISIZE(src), kernel, apertureSize, avgWindow, buffer.getWriteData());
	}

void FastMarchingGetBufferSize(const QVGenericImage &image, QVImage<uChar> &buffer)
	{
	int bufferSize;
	IPP_DEBUG(CannyGetSize, ippiFastMarchingGetBufferSize_8u32f_C1R, IMAGE_ROISIZE(image),&bufferSize);
	buffer = QVImage<uChar>(bufferSize,1);
	}

void FastMarching(const QVImage<uChar> &mask, QVImage<sFloat> &distances, int radius, QVImage<uChar> &buffer, QPoint destROIOffset)
	{
	distances.setROI(distances.getAnchor().x(), distances.getAnchor().y(), mask.getROI().width(), mask.getROI().height());

	IPP_DEBUG(FastMarching,	ippiFastMarching_8u32f_C1R,
				PDATA_READ(mask), mask.getStep(),
				PDATA_WRITE(distances), distances.getStep(),
				IMAGE_ROISIZE(mask), radius, buffer.getWriteData());
	}

void Inpaint(const QVImage<uChar> &src, const QVImage<uChar> &mask, const QVImage<sFloat> &distances,
		QVImage<uChar> &dest, int radius,  IppiInpaintFlag flag, QPoint destROIOffset)
	{
	IppiInpaintState_8u_C1R * pState;

	IPP_DEBUG(	Inpaint, ippiInpaintInitAlloc_8u_C1R, &pState, distances.getReadData(), distances.getStep(),
			mask.getReadData(), mask.getStep(), IMAGE_ROISIZE(mask), radius, flag);

	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());

	IPP_DEBUG(Inpaint,	ippiInpaint_8u_C1R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(dest), dest.getStep(),
				IMAGE_ROISIZE(src), pState);

	IPP_DEBUG(Inpaint, ippiInpaintFree_8u_C1R,pState);
	}

void Inpaint(const QVImage<uChar,3> &src, const QVImage<uChar> &mask, const QVImage<sFloat> &distances,
		QVImage<uChar,3> &dest, int radius,  IppiInpaintFlag flag, QPoint destROIOffset)
	{
	IppiInpaintState_8u_C3R * pState;

	IPP_DEBUG(	Inpaint, ippiInpaintInitAlloc_8u_C3R, &pState, distances.getReadData(), distances.getStep(),
			mask.getReadData(), mask.getStep(), IMAGE_ROISIZE(mask), radius, flag);

	dest.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width(), src.getROI().height());

	IPP_DEBUG(Inpaint,	ippiInpaint_8u_C3R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(dest), dest.getStep(),
				IMAGE_ROISIZE(src), pState);

	IPP_DEBUG(Inpaint, ippiInpaintFree_8u_C3R,pState);
	}

QVector<int> HistogramRange(const QVImage<uChar, 1> &src)
	{
	// Bug: always coputes last value wrong. So we use one more value, 257, instead of 256.
	int nLevels = 257;
	sInt pLevels[nLevels];
	int pHist[nLevels];
	
	for(sInt i=0; i<nLevels; i++)
		pLevels[i] = i;

	IPP_DEBUG(HistogramRange,	ippiHistogramRange_8u_C1R, // ippi ## NAME ## SUBFIX,
					PDATA_READ(src), src.getStep(), IMAGE_ROISIZE(src),
					pHist, pLevels, nLevels);

	////////////////////
	// compute cumulative histogram
	QVector< int > result(256);
	for (int i=0; i<result.size(); i++)
		result[i] = pHist[i];

	return result;
	}

QMap<sInt, int> HistogramRange(const QVImage<uChar, 1> &src, QList<uChar> values)
	{
	// Bug: always coputes last value wrong. So we use one more value, 257, instead of 256.
	int nLevels = values.size();
	sInt pLevels[nLevels];
	int pHist[nLevels];
	
	Q_ASSERT_X(false, "QMap<sInt, int> HistogramRange(QVImage<uChar, 1> &src, QList<uChar> values)",
			"Tried an user defined values QList. Functionality not implemented yet.");

	IPP_DEBUG(HistogramRange,	ippiHistogramRange_8u_C1R, // ippi ## NAME ## SUBFIX,
					PDATA_READ(src), src.getStep(), IMAGE_ROISIZE(src),
					pHist, pLevels, nLevels);

	////////////////////
	// compute cumulative histogram
	QMap<sInt, int> histogram;
	for (int i=0; i<nLevels; i++)
		histogram.insert(pLevels[i], pHist[i]);

	return histogram;
	}

// Flood fill
void	FloodFillGetSize(const QVGenericImage &src, QVImage<uChar> &buffer)
	{
	int floodFillBufferSize;
	IPP_DEBUG(FloodFillGetSize, ippiFloodFillGetSize, IMAGE_ROISIZE(src),&floodFillBufferSize);
	buffer = QVImage<uChar>(floodFillBufferSize,1);
	}

#define CREATE_IPP_FUNCTION_FLOOD_FILL(NAME, CONNECTIVITY, SUBFIX, TYPE)		\
void NAME(QVImage<TYPE> &img, QVImage<uChar> &buffer,					\
	uInt seedX, uInt seedY, TYPE newVal, TYPE minDelta, TYPE maxDelta)		\
	{										\
	IppiConnectedComp pRegion;							\
	IPP_DEBUG(NAME,		ippiFloodFill ## CONNECTIVITY ## SUBFIX,		\
				PDATA_WRITE(img), img.getStep(), IMAGE_ROISIZE(img),	\
				IPP_POINT(seedX, seedY), newVal, minDelta, maxDelta,	\
				&pRegion, buffer.getWriteData());			\
	}

CREATE_IPP_FUNCTION_FLOOD_FILL(FloodFillRange4Connected, _Range4Con, _8u_C1IR, uChar);
CREATE_IPP_FUNCTION_FLOOD_FILL(FloodFillGradient4Connected, _Grad4Con, _8u_C1IR, uChar);

#define CREATE_IPP_FUNCTION_WARP_PERSPECTIVE(NAME, SUBFIX, C, TYPE)									\
void	NAME(const QVImage<TYPE, C> &src, QVImage<TYPE, C> &dest, const QVMatrix &rectifyingHomography, int interpolation, QPoint destROIOffset)	\
	{																		\
	double coeffs[3][3];																\
	for (int i = 0; i < 3; i++)															\
		for (int j = 0; j < 3; j++)														\
			coeffs[i][j] = rectifyingHomography(i,j);											\
																			\
	IPP_DEBUG(NAME,		ippi ## NAME ## SUBFIX,													\
				src.getReadData(), IMAGE_SIZE(src), src.getStep(), IMAGE_ROIRECT(src),							\
				PDATA_WRITE(dest), dest.getStep(), IMAGE_ROIRECT(dest),									\
				coeffs, interpolation);													\
	}

CREATE_IPP_FUNCTION_WARP_PERSPECTIVE(WarpPerspective, _8u_C1R, 1, uChar);
CREATE_IPP_FUNCTION_WARP_PERSPECTIVE(WarpPerspective, _8u_C3R, 3, uChar);

#include <qvmath/qvmatrix.h>
void	Filter(const QVImage<sFloat> &src, QVImage<sFloat> &dest, const QVMatrix kernel, const QPoint destROIOffset)
	{
	Q_ASSERT(kernel.getCols() == kernel.getRows());

	const int kernelSize = kernel.getCols();
	sFloat kernelValues[kernelSize*kernelSize];
	for(int i = 0; i < kernelSize; i++)
		for(int j = 0; j < kernelSize; j++)
			kernelValues[i+ kernelSize*j] = kernel(i,j);

	dest.setROI(destROIOffset.x(), destROIOffset.y(),
		src.getROI().width()-kernelSize+1, src.getROI().height()-kernelSize+1);

	/// @note ojo, quizás haya que indicar en IPP_POINT(0,0), el último elemento de la matriz 'kernel', tal y como pasa en las funciones
	/// FilterColumn y FilterRow.
	IPP_DEBUG(NAME,		ippiFilter_32f_C1R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(dest), dest.getStep(),
				IMAGE_ROISIZE(dest),
				kernelValues, IPP_SIZE(kernelSize, kernelSize), IPP_POINT(0,0)
				);
	}

/// @note Ojo: muerte, peligro. Hay que indicar como anchor, el tamaño del kernel. La iteración irá desde el valor 'kernelSize', hacia abajo.
void	FilterColumn(const QVImage<sFloat> &src, QVImage<sFloat> &dest, const QVVector kernel, const QPoint destROIOffset)
	{
	const int kernelSize = kernel.size();
	sFloat kernelValues[kernelSize];
	for(int i = 0; i < kernelSize; i++)
			kernelValues[i] = kernel[i];

	dest.setROI(destROIOffset.x(), destROIOffset.y()+kernelSize,
		src.getROI().width(), src.getROI().height()-kernelSize+1);

	IPP_DEBUG(NAME,		ippiFilterColumn_32f_C1R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(dest), dest.getStep(),
				IMAGE_ROISIZE(dest),
				kernelValues, kernelSize, kernelSize
				);
	}

/// @note Ojo: muerte, peligro. Hay que indicar como anchor, el tamaño del kernel. La iteración irá desde el valor 'kernelSize', hacia abajo.
void	FilterRow(const QVImage<sFloat> &src, QVImage<sFloat> &dest, const QVVector kernel, const QPoint destROIOffset)
	{
	const int kernelSize = kernel.size();
	sFloat kernelValues[kernelSize];
	for(int i = 0; i < kernelSize; i++)
			kernelValues[i] = kernel[i];

	dest.setROI(destROIOffset.x()+kernelSize, destROIOffset.y(),
		src.getROI().width()-kernelSize+1, src.getROI().height());

	IPP_DEBUG(NAME,		ippiFilterRow_32f_C1R,
				PDATA_READ(src), src.getStep(),
				PDATA_WRITE(dest), dest.getStep(),
				IMAGE_ROISIZE(dest),
				kernelValues, kernelSize, kernelSize
				);
	}

void SqrIntegral(const QVImage<uChar,1> &src, QVImage<sFloat,1> &dst1, QVImageBuffer<sDouble,1> &dst2, sFloat val, sDouble valSqr, QPoint destROIOffset)
{
	dst1.setROI(destROIOffset.x(), destROIOffset.y(), src.getROI().width()+1, src.getROI().height()+1);

	IPP_DEBUG(SqrIntegral,ippiSqrIntegral_8u32f64f_C1R,
			PDATA_READ(src), src.getStep(),
			PDATA_WRITE(dst1), dst1.getStep(),
			dst2.getWriteData(), dst2.getStep(),
			IMAGE_ROISIZE(src), val,valSqr);

}


void RectStdDev(const QVImage<sFloat,1> &src1, const QVImageBuffer<sDouble,1> &src2, QVImage<sFloat,1> &dst, IppiRect rect, QPoint destROIOffset)
{
	dst.setROI(destROIOffset.x(), destROIOffset.y(), src1.getROI().width()-rect.width, src1.getROI().height()-rect.height);

	IPP_DEBUG(RectStdDev,ippiRectStdDev_32f_C1R,
			PDATA_READ(src1), src1.getStep(),
			src2.getReadData(), src2.getStep(),
			PDATA_WRITE(dst), dst.getStep(),
			IMAGE_ROISIZE(dst), rect);

	//OJO
	dst.resetROI();

}
