/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvquaternion.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVQUATERNION_H
#define QVQUATERNION_H

#include <math.h>
#include <iostream>
#include <qvmath/qvvector.h>

class QVMatrix;

/*!
@class QVQuaternion qvquaternion.h qvmath/qvquaternion.h
@brief Implementation of <a href="http://en.wikipedia.org/wiki/Quaternion">quaternions</a>.

@todo develop notes about quaternions.

@ingroup qvmath
*/

class QVQuaternion: public QVVector
	{
	public:
		//////////////////
		// Constructors

		/// @brief Identity constructor.
		///
		/// Creates quaternion to identity rotation.
		QVQuaternion();

		/// @brief Axis rotation constructor.
		///
		/// Creates a quaternion object which rotates around a direction a given angle.
		///
		/// @param direction quaternion will rotate around this direction.
		/// @param phi quaternion will rotate this angle, given in radians.
		QVQuaternion(QVVector direction, float phi);

		/// @brief Direct coordinates constructor.
		///
		/// Creates a quaternion from given coordinates.
		///
		/// @param q1 first coordinate for quaternion.
		/// @param q2 second coordinate for quaternion.
		/// @param q3 third coordinate for quaternion.
		/// @param q4 fourth coordinate for quaternion.
		QVQuaternion(const double q1, const double q2, const double q3, const double q4);

		/// @brief Rotation matrix constructor
		///
		/// Creates a quaternion from a rotation matrix.
		///
		/// @param matrix rotation matrix to create the quaternion from
		QVQuaternion(const QVMatrix matrix);

		//////////////////////////

		/// @brief Quaternion product
		///
		/// @param quaternion operand for the product
		/// @see quaternionProduct
		QVQuaternion operator*(const QVQuaternion &quaternion) const		{ return quaternionProduct(quaternion); 	};

		/// @brief Copy constructor.
		///
		/// Creates a quaternion copying its coordinates from a given one.
		///
		/// @param quaternion quaternion to copy coordinates
		QVQuaternion(const QVQuaternion &quaternion): QVVector(quaternion)	{}

		/// @brief Gets quaternion corresponding to a trackball rotation
		///
		/// Creates a quaternion which rotates given two points at a plane, mapped to a sphere.
		///
		/// @param p1x first coordinate for first point.
		/// @param p1y second coordinate for first point.
		/// @param p2x first coordinate for second point.
		/// @param p2y second coordinate for second point.
		/// @returns quaternion corresponding to rotate in a sphere from point (p1x, p1y) to (p2x, p2y).
		/// corresponding to this quaternion, and at element (4,4), the unity.
		static QVQuaternion trackball(float p1x, float p1y, float p2x, float p2y);

		/// @brief Combine operation for quaternions
		///
		/// Creates a quaternion corresponding to combination of rotations corresponding to actual
		/// quaternion and the one received by parameter
		///
		/// @param quaternion operand in the combining operation.
		/// @returns resulting quaternion.
		QVQuaternion quaternionProduct(const QVQuaternion &quaternion) const;

	private:
		QVQuaternion normalizeQuaternion() const;
	};

std::ostream& operator << ( std::ostream &os, const QVQuaternion &quaternion );

#endif

