/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvvector.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVVECTOR_H
#define QVVECTOR_H

#include <math.h>
#include <iostream>
#include <QVector>
#include <QPointF>
#include <qvcore/qvdefines.h>
#include <gsl/gsl_linalg.h>

class QVMatrix;

/*!
@class QVVector qvvector.h qvmath/qvvector.h
@brief Implementation of numerical vectors.

@ingroup qvmath
*/
class QVVector: public QVector<double>
	{
	public:
		/// @brief Default constructor for QVVector
		///
		/// @param vector operand for the product
		QVVector(): QVector<double>()									{}

		/// @brief Constructor for QVVector
		///
		/// @param int initializing size for the vector
		/// @param defaultValue default value to set all elements of the vector
		QVVector(const int size, const double defaultValue = 0): QVector<double>(size, defaultValue)	{}

		/// @brief Copy constructor for QVVector
		///
		/// @param vector vector to be copied
		QVVector(const QVVector &vector): QVector<double>(vector)					{}

		/// @brief Copy constructor for QVector
		///
		/// @param vector vector to be copied
		QVVector(const QVector<double> &vector): QVector<double>(vector)				{}

		/// @brief Convert constructor from QVMatrix
		///
		/// This function converts a QVMatrix object into a QVVector object.
		/// Given a matrix \$ A_{m,n} \$, this function generates a row vector of size mxn, containing
		/// each one of the rows of the original matrix.
		///
		/// @param matrix matrix to be converted.
		QVVector(const QVMatrix &matrix);

		/// @brief Copy constructor for gsl_vector
		///
		/// @param vector vector to be copied
		QVVector(const gsl_vector *vector): QVector<double>(vector->size)
				{
				for(int i = 0; i < size(); i++)
					replace(i, gsl_vector_get(vector, i));
				}

		/// @brief Convert constructor for QPoint
		///
		/// @param vector vector to be copied
		QVVector(const QPoint &point): QVector<double>(2)	{ operator[](0) = point.x(); operator[](1) = point.y(); }

		/// @brief Convert constructor for QPointF
		///
		/// @param vector vector to be copied
		QVVector(const QPointF &point): QVector<double>(2)	{ operator[](0) = point.x(); operator[](1) = point.y(); }

		///////////////////////////////

		/// @brief Vector dot product
		///
		/// @param vector operand for the product
		/// @see dotProduct
		double operator*(const QVVector &vector) const		{ return dotProduct(vector); };

		/// @brief Vector cross product
		///
		/// @param vector operand for the product
		/// @see crossProduct
		QVVector operator^(const QVVector &vector) const	{ return crossProduct(vector); };

		/// @brief Vector add
		///
		/// @param vector term for the operation
		/// @see add
		QVVector operator+(const QVVector &vector) const	{ return add(vector); };

		/// @brief Vector substraction
		///
		/// @param vector subtrahend for the operation
		/// @see substract
		QVVector operator-(const QVVector &vector) const	{ return substract(vector); };

		/// @brief Scalar multiplication
		/// @todo operator* para matrics. Devuelve un vector, resultado de multiplicar el vector actual por la matriz, por el lado izquierdo.
		///
		/// @param value value to multiply each component of the vector
		/// @see operator*(const double)
		QVVector operator*(const double value) const		{ return scalarMultiplication(value); };

		/// @brief Scalar division
		///
		/// @param value value to divide each component of the vector
		/// @see scalarDivide
		QVVector operator/(const double value) const		{ return scalarDivision(value); };

		/// @brief Vector-matrix product
		///
		/// @param matrix operand for the product
		QVVector operator*(const QVMatrix &matrix) const;

		/// @brief Compare operator
		///
		/// @param vector vector to compare
		/// @see equals
		bool operator==(const QVVector &vector) const		{ return equals(vector); };

		/// @brief Add compound assignment operator
		///
		/// @param vector vector to add
		QVVector & operator+=(const QVVector &vector)		{ return (*this = add(vector));	}


		/// @brief Convert to QPointF operator
		///
		/// Cast from homogeneous coordinates.
		///
		operator QPointF() const
			{
			Q_ASSERT(size() > 1);
			Q_ASSERT(size() < 4);
			if (size() == 2)
				return QPointF(operator[](0), operator[](1));
			else
				return QPointF(operator[](0)/operator[](2), operator[](1)/operator[](2));
			}

		/// @brief Convert to gsl_vector * operator
		///
		/// @param value gsl_vector
		operator gsl_vector * () const
			{
			gsl_vector *result = gsl_vector_alloc(size());
			const double *data = this->data();
			for(uint i = 0; i < result->size; i++)
				gsl_vector_set (result, i, data[i]);
			return result;
			}

		////////////////

		/// @brief Set vector elements to a value
		///
		/// @param value value to set all the elements of the vector
		void set(const double value)
			{
			for (int i = 0; i < size(); i++)
				operator[](i) =  value;
			};

		/// @brief Set vector elements to a value
		///
		/// @param value value to set all the elements of the vector
		QVVector subVector(const int firstIndex, const int lastIndex)
			{
			Q_ASSERT(0 <= firstIndex);
			Q_ASSERT(firstIndex <= lastIndex);
			Q_ASSERT(lastIndex < this->size());

			QVVector result(lastIndex - firstIndex +1);
			for (int i = firstIndex; i <= lastIndex; i++)
				result[i] = operator[](i);

			return result;
			};

		/// @brief Scalar division
		///
		/// @param value value to divide each component of the vector
		/// @see operator/(const double)
		QVVector scalarDivision(const double value) const
			{
			QVVector result = *this;
			for (int i = 0; i < size(); i++)
				result[i] /= value;
			return result;
			};

		/// @brief Scalar multiplication
		///
		/// @param value value to multiply each component of the vector
		/// @see operator*(const double)
		QVVector scalarMultiplication(const double value) const
			{
			QVVector result = *this;
			for (int i = 0; i < size(); i++)
				result[i] *= value;
			return result;
			};

		/// @brief Norm 2 of the vector
		///
		/// @see normalize
		double norm2() const
			{ return sqrt(*this * *this); }

		/// @brief Normalize vector
		///
		/// Divides each element of the vector by its second norm.
		/// @see norm
		QVVector normalize()	const
			{ return operator/(norm2()); }

		/// @brief Vector dot product
		///
		/// @param vector operand for the product
		/// @see operator*
		double dotProduct(const QVVector &vector) const;

		/// @brief Vector cross product
		///
		/// Computes the cross product for two vectors of size 3
		/// @param vector operand for the product
		/// @see operator^
		QVVector crossProduct(const QVVector &vector) const;

		/// @brief Vector add
		///
		/// @param vector term for the operation
		/// @see operator+
		QVVector add(const QVVector &vector) const;

		/// @brief Vector substraction
		///
		/// @param vector subtrahend for the operation
		/// @see operator-
		QVVector substract(const QVVector &vector) const;

		/// @brief Compare operator
		///
		/// @param vector vector to compare
		/// @see equals
		bool equals(const QVVector &vector) const;

		///////////////////////////

		/// @brief Gaussian distributed values vector.
		///
		/// This function generates a vector containing values for a gaussian distribution, with median in the middle value of the
		/// vector, and deviation indicated.
		///
		/// @param vector vector to compare
		/// @see equals
		static const QVVector gaussianDistribution(const int size, const double sigma)
			{
			const double nu = ((double) size)/2;
			QVVector result(size);
			for(int x=0; x< size; x++)
				result[x] = exp(-POW2(x+0.5-nu)/(2*POW2(sigma))) / ( sigma * sqrt(2*PI));
		
			return result;
			}
	};

/// @todo detailed description should follow
/// @ingroup qvmath
std::ostream& operator << ( std::ostream &os, const QVVector &vector );

/// @todo detailed description should follow
/// @ingroup qvmath
uint qHash(const QVVector &vector);

#endif

